# This document is free and open-source software, subject to the OSI-approved
# BSD license below.
#
# Copyright (c) 2014 Alexis Petrounias <www.petrounias.org>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the author nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Unit tests for collections
"""
from collections import OrderedDict, defaultdict

import jsonpickle

__status__ = "Stable"
__version__ = "1.0.0"
__maintainer__ = "Alexis Petrounias <www.petrounias.org>"
__author__ = "Alexis Petrounias <www.petrounias.org>"


"""
Classes featuring various collections used by
:mod:`restorable_collections_tests` unit tests; these classes must be
module-level (including the default dictionary factory method) so that they
can be pickled.

Tests restorable collections by creating pickled structures featuring
no cycles, self cycles, and mutual cycles, for all supported dictionary and
set wrappers. Python built-in dictionaries and sets are also tested with
expectation of failure via raising exceptions.

"""


class Group:
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.elements = []

    def __repr__(self):
        return f'Group({self.name})'


class C:
    def __init__(self, v):
        super().__init__()
        self.v = v
        self.plain = dict()
        self.plain_ordered = OrderedDict()
        self.plain_default = defaultdict(c_factory)

    def add(self, key, value):
        self.plain[key] = (key, value)
        self.plain_ordered[key] = (key, value)
        self.plain_default[key] = (key, value)

    def __hash__(self):
        return hash(self.v) if hasattr(self, 'v') else id(self)

    def __repr__(self):
        return f'C({self.v})'


def c_factory():
    return (C(0), '_')


class D:
    def __init__(self, v):
        super().__init__()
        self.v = v
        self.plain = set()

    def add(self, item):
        self.plain.add(item)

    def __hash__(self):
        return hash(self.v) if hasattr(self, 'v') else id(self)

    def __repr__(self):
        return f'D({self.v})'


def pickle_and_unpickle(obj):
    encoded = jsonpickle.encode(obj, keys=True)
    return jsonpickle.decode(encoded, keys=True)


def test_dict_no_cycle():
    g = Group('group')
    c1 = C(42)
    g.elements.append(c1)
    c2 = C(67)
    g.elements.append(c2)
    c1.add(c2, 'a')  # points to c2, which does not point to anything

    assert c2 in c1.plain
    assert c2 in c1.plain_ordered
    assert c2 in c1.plain_default

    gu = pickle_and_unpickle(g)
    c1u = gu.elements[0]
    c2u = gu.elements[1]

    # check existence of keys directly
    assert c2u in c1u.plain.keys()
    assert c2u in c1u.plain_ordered.keys()
    assert c2u in c1u.plain_default.keys()

    # check direct key-based lookup
    assert c2u == c1u.plain[c2u][0]
    assert c2u == c1u.plain_ordered[c2u][0]
    assert c2u == c1u.plain_default[c2u][0]

    # check key lookup with key directly from keys()
    plain_keys = list(c1u.plain.keys())
    ordered_keys = list(c1u.plain_ordered.keys())
    default_keys = list(c1u.plain_default.keys())
    assert c2u == c1u.plain[plain_keys[0]][0]
    assert c2u == c1u.plain_ordered[ordered_keys[0]][0]
    assert c2u == c1u.plain_default[default_keys[0]][0]
    assert c2u == c1u.plain[plain_keys[0]][0]
    assert c2u == c1u.plain_ordered[ordered_keys[0]][0]
    assert c2u == c1u.plain_default[default_keys[0]][0]


def test_dict_self_cycle():
    g = Group('group')
    c1 = C(42)
    g.elements.append(c1)
    c2 = C(67)
    g.elements.append(c2)
    c1.add(c1, 'a')  # cycle to itself
    c1.add(c2, 'b')  # c2 does not point to itself nor c1

    assert c1 in c1.plain
    assert c1 in c1.plain_ordered
    assert c1 in c1.plain_default

    gu = pickle_and_unpickle(g)
    c1u = gu.elements[0]
    c2u = gu.elements[1]

    # check existence of keys directly
    # key c1u
    assert c1u in list(c1u.plain.keys())
    assert c1u in list(c1u.plain_ordered.keys())
    assert c1u in list(c1u.plain_default.keys())

    # key c2u
    assert c2u in list(c1u.plain.keys())
    assert c2u in list(c1u.plain_ordered.keys())
    assert c2u in list(c1u.plain_default.keys())

    # check direct key-based lookup

    # key c1u
    assert 42 == c1u.plain[c1u][0].v
    assert 42 == c1u.plain_ordered[c1u][0].v

    assert len(c1u.plain_default) == 2
    assert 42 == c1u.plain_default[c1u][0].v
    # No new entries were created.
    assert len(c1u.plain_default) == 2

    # key c2u
    # succeeds because c2u does not have a cycle to itself
    assert c2u == c1u.plain[c2u][0]
    # succeeds because c2u does not have a cycle to itself
    assert c2u == c1u.plain_ordered[c2u][0]
    # succeeds because c2u does not have a cycle to itself
    assert c2u == c1u.plain_default[c2u][0]
    assert c2u == c1u.plain[c2u][0]
    assert c2u == c1u.plain_ordered[c2u][0]
    assert c2u == c1u.plain_default[c2u][0]

    # check key lookup with key directly from keys()
    # key c1u
    plain_keys = list(c1u.plain.keys())
    ordered_keys = list(c1u.plain_ordered.keys())
    default_keys = list(c1u.plain_default.keys())
    value = 42
    assert value == c1u.plain[plain_keys[0]][0].v
    42 == c1u.plain_ordered[ordered_keys[0]][0].v
    42 == c1u.plain_default[default_keys[0]][0].v

    # key c2u
    # succeeds because c2u does not have a cycle to itself
    assert c2u == c1u.plain[plain_keys[1]][0]
    # succeeds because c2u does not have a cycle to itself
    assert c2u == c1u.plain_ordered[ordered_keys[1]][0]
    assert 67 == c1u.plain_default[default_keys[1]][0].v


def test_dict_mutual_cycle():
    g = Group('group')
    c1 = C(42)
    g.elements.append(c1)
    c2 = C(67)
    g.elements.append(c2)

    c1.add(c2, 'a')  # points to c2, which points to c1, forming cycle
    c2.add(c1, 'a')  # points to c1 in order to form cycle

    assert c2 in c1.plain
    assert c2 in c1.plain_ordered
    assert c2 in c1.plain_default

    assert c1 in c2.plain
    assert c1 in c2.plain_ordered
    assert c1 in c2.plain_default

    gu = pickle_and_unpickle(g)
    c1u = gu.elements[0]
    c2u = gu.elements[1]

    # check existence of keys directly
    # key c2u
    assert c2u in c1u.plain.keys()
    assert c2u in c1u.plain_ordered.keys()
    assert c2u in c1u.plain_default.keys()

    # key c1u
    assert c1u in list(c2u.plain.keys())
    assert c1u in list(c2u.plain_ordered.keys())
    assert c1u in list(c2u.plain_default.keys())

    # check direct key-based lookup
    # key c2u, succeed because c2u added to c1u after __setstate__
    assert c2u == c1u.plain[c2u][0]
    assert c2u == c1u.plain_ordered[c2u][0]
    assert c2u == c1u.plain_default[c2u][0]

    # key c1u
    assert isinstance(c2u.plain[c1u][0], C)
    assert 42 == c2u.plain[c1u][0].v
    assert isinstance(c2u.plain_ordered[c1u][0], C)

    # succeeds
    assert len(c2u.plain_default) == 1
    assert 42 == c2u.plain_default[c1u][0].v
    # Ensure that no new entry is created by verifying that the length
    # has not changed.
    assert len(c2u.plain_default) == 1

    # check key lookup with key directly from keys()

    # key c2u, succeed because c2u added to c1u after __setstate__
    plain_keys = list(c1u.plain.keys())
    ordered_keys = list(c1u.plain_ordered.keys())
    default_keys = list(c1u.plain_default.keys())
    assert c2u == c1u.plain[plain_keys[0]][0]
    assert c2u == c1u.plain_ordered[ordered_keys[0]][0]
    assert c2u == c1u.plain_default[default_keys[0]][0]

    # key c1u
    plain_keys = list(c2u.plain.keys())
    ordered_keys = list(c2u.plain_ordered.keys())
    default_keys = list(c2u.plain_default.keys())
    assert 42 == c2u.plain[plain_keys[0]][0].v
    assert 42 == c2u.plain_ordered[ordered_keys[0]][0].v
    assert 42 == c2u.plain_default[default_keys[0]][0].v
    assert isinstance(c2u.plain[plain_keys[0]], tuple)
    assert isinstance(c2u.plain_ordered[ordered_keys[0]], tuple)


def test_set_no_cycle():
    g = Group('group')
    d1 = D(42)
    g.elements.append(d1)
    d2 = D(67)
    g.elements.append(d2)
    d1.add(d2)  # d1 points to d1, d2 does not point to anything, no cycles

    assert d2 in d1.plain

    gu = pickle_and_unpickle(g)
    d1u = gu.elements[0]
    d2u = gu.elements[1]

    # check element directly
    assert d2u in d1u.plain
    # check element taken from elements
    assert list(d1u.plain)[0] in d1u.plain


def test_set_self_cycle():
    g = Group('group')
    d1 = D(42)
    g.elements.append(d1)
    d2 = D(67)
    g.elements.append(d2)
    d1.add(d1)  # cycle to itself
    d1.add(d2)  # d1 also points to d2, but d2 does not point to d1

    assert d1 in d1.plain

    gu = pickle_and_unpickle(g)
    d1u = gu.elements[0]
    d2u = gu.elements[1]
    assert d2u is not None

    # check element directly
    assert d1u in d1u.plain
    # check element taken from elements
    assert list(d1u.plain)[0] in d1u.plain
    # succeeds because d2u added to d1u after __setstate__
    assert list(d1u.plain)[1] in d1u.plain


def test_set_mutual_cycle():
    g = Group('group')
    d1 = D(42)
    g.elements.append(d1)
    d2 = D(67)
    g.elements.append(d2)
    d1.add(d2)  # points to d2, which points to d1, forming cycle
    d2.add(d1)  # points to d1 in order to form cycle

    assert d2 in d1.plain
    assert d1 in d2.plain

    gu = pickle_and_unpickle(g)
    d1u = gu.elements[0]
    d2u = gu.elements[1]

    # check element directly
    # succeeds because d2u added to d1u after __setstate__
    assert d2u in d1u.plain
    assert d1u in d2u.plain
    # check element taken from elements
    # succeeds because d2u added to d1u after __setstate__
    assert list(d1u.plain)[0] in d1u.plain
    assert list(d2u.plain)[0] in d2u.plain



import traceback  # pragma: no cover
from typing import Dict, List, Tuple, Union  # pragma: no cover

import atheris  # pragma: no cover


@atheris.instrument_func
def is_expected_error(
    exception: Exception, expected_errors: Dict[str, List[Tuple[str, Union[int]]]]
):  # pragma: no cover
    """Checks if a given exception matches any of the expected errors.

    This function inspects the traceback of the provided exception to determine if
    it originated from a file and line number listed in the expected errors, and if
    the exception message contains a specified substring.

    Args:
        exception (Exception): The exception to be checked.
        expected_errors (Dict[str, List[Tuple[str, Union[int]]]]): A dictionary where
            keys are filenames and values are lists of tuples. Each tuple contains a
            substring of an expected exception message and a line number (or -1 to
            match any line number).

    Returns:
        bool: True if the exception matches any of the expected errors, False otherwise.
    """
    tb = traceback.extract_tb(exception.__traceback__)
    if not tb:
        return False

    last_frame = tb[-1]
    exception_origin_file = last_frame.filename

    if exception_origin_file in expected_errors:
        origin_line = int(last_frame.lineno)
        exception_message = str(exception).lower()
        for message, line in expected_errors[exception_origin_file]:
            if message.lower() in exception_message and (
                line == -1 or line == origin_line
            ):
                return True
    return False



"""Test serializing sqlalchemy models"""

from helper import SkippableTest

import jsonpickle

try:
    import sqlalchemy as sqa
    from sqlalchemy import orm
    from sqlalchemy.ext import declarative
    from sqlalchemy.orm import Session

    HAS_SQA = True
except ImportError:
    HAS_SQA = False

if HAS_SQA:
    # sqlalchemy.ext.declarative.declarative_base() was deprecated in SQLAlchemy 2.0
    # and replaced by sqlalchemy.orm.declarative_base().
    if hasattr(orm, 'declarative_base'):
        Base = orm.declarative_base()
    else:
        Base = declarative.declarative_base()

    class Table(Base):
        __tablename__ = 'table'
        id = sqa.Column(sqa.Integer, primary_key=True)
        name = sqa.Column(sqa.Text)
        value = sqa.Column(sqa.Float)


class SQLAlchemyTestCase(SkippableTest):

    def setUp(self):
        """Create a new sqlalchemy engine for the test"""
        if HAS_SQA:
            url = 'sqlite:///:memory:'
            self.engine = sqa.create_engine(url)
            Base.metadata.drop_all(self.engine)
            Base.metadata.create_all(self.engine)
            self.should_skip = False
        else:
            self.should_skip = True

    def test_sqlalchemy_roundtrip_with_detached_session(self):
        """Test cloned SQLAlchemy objects detached from any session"""
        if self.should_skip:
            return self.skip('sqlalchemy is not installed')
        expect = Table(name='coolness', value=11.0)
        session = Session(bind=self.engine, expire_on_commit=False)
        session.add(expect)
        session.commit()
        jsonstr = jsonpickle.dumps(expect)
        actual = jsonpickle.loads(jsonstr)
        # actual is a shadow object; it cannot be added to the same
        # session otherwise sqlalchemy will detect an identity conflict.
        # To make this work we use expire_on_commit=True so that sqlalchemy
        # allows us to do read-only operations detached from any session.
        assert expect.id == actual.id
        assert expect.name == actual.name
        assert expect.value == actual.value

    def test_sqlalchemy_roundtrip_with_two_sessions(self):
        """Test cloned SQLAlchemy objects attached to a secondary session"""
        if self.should_skip:
            return self.skip('sqlalchemy is not installed')
        expect = Table(name='coolness', value=11.0)
        session = Session(bind=self.engine, expire_on_commit=False)
        session.add(expect)
        session.commit()
        jsonstr = jsonpickle.dumps(expect)
        actual = jsonpickle.loads(jsonstr)
        # actual is a shadow object; it cannot be added to the same
        # session otherwise sqlalchemy will detect an identity conflict.
        # To make this work we use expire_on_commit=True so that sqlalchemy
        # allows us to do read-only operations detached from any session.
        assert expect.id == actual.id
        assert expect.name == actual.name
        assert expect.value == actual.value

    def test_sqlalchemy_with_dynamic_table(self):
        """Test creating a table dynamically, per #180"""
        if self.should_skip:
            return self.skip('sqlalchemy is not installed')
        meta = sqa.MetaData()
        expect = sqa.Table(
            'test',
            meta,
            sqa.Column('id', sqa.Integer()),
            sqa.Column('text', sqa.Text()),
        )
        jsonstr = jsonpickle.dumps(expect)
        actual = jsonpickle.loads(jsonstr)
        assert expect.__class__ == actual.__class__
        assert expect.name == actual.name
        # These must be unique instances
        assert expect.metadata != actual.metadata
        # Columns names must exactly match
        assert sorted(expect.columns.keys()) == sorted(actual.columns.keys())
        # As should the types
        assert expect.c.id.name == actual.c.id.name
        assert expect.c.id.type.__class__ == actual.c.id.type.__class__
        assert expect.c.text.name == actual.c.text.name
        assert expect.c.text.type.__class__ == actual.c.text.type.__class__



#!/usr/bin/env python
#
# Copyright (C) 2008 John Paulett (john -at- paulett.org)
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

import argparse
import os
import sys
import timeit

json = """\
import feedparser
import jsonpickle
import feedparser_test as test
doc = feedparser.parse(test.RSS_DOC)

jsonpickle.set_preferred_backend('%s')

pickled = jsonpickle.encode(doc)
unpickled = jsonpickle.decode(pickled)
if doc['feed']['title'] != unpickled['feed']['title']:
    print('Not a match')
"""


if __name__ == '__main__':
    sys.path.insert(1, os.getcwd())

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-n',
        '--number',
        type=int,
        default=500,
        help='number of iterations to benchmark',
    )
    parser.add_argument('mod', nargs='?', default='json', help='json module')

    args = parser.parse_args()
    mod = args.mod
    number = args.number

    print('Using %s' % mod)
    json_test = timeit.Timer(stmt=json % mod)
    print("%.9f sec/pass " % (json_test.timeit(number=number) / number))



"""
Run "make benchmark" in the directory that this file is in.
This will test the version of jsonpickle inside the local directory,
NOT the pip installed version.
To test the pip installed version, run this outside the jsonpickle directory:
py.test --benchmark-only --benchmark-histogram=./jsonpickle/images/benchmark-$(date +%Y-%m-%dT%T%z) ./jsonpickle/jsonpickle_benchmarks.py
MAKE SURE you have pytest-benchmark pip installed.
"""

import itertools

import jsonpickle

# DEFINTIIONS:
# HOMOGENOUS - Except for sets, each container may only have integers, floats, strings,
# and itself inside it
# HETEROGENOUS - Except for sets, each container must have at least one of every other
# container type being tested
# SIMPLE - Each container must have two each of integers, floats, and strings
# COMPLEX - Each container must have four each of integers, floats, and strings

# SETUP
###################################################################################


class SlotPickleMixin(object):
    def __getstate__(self):
        all_slots = itertools.chain.from_iterable(
            getattr(cls, "__slots__", []) for cls in self.__class__.__mro__
        )
        return dict(
            (slot, getattr(self, slot)) for slot in all_slots if hasattr(self, slot)
        )

    def __setstate__(self, state):
        for slot, value in dict(state).items():
            setattr(self, slot, value)


class MyClass(object):
    __slots__ = ["idk", "idk2"]

    def __init__(self):
        self.idk = {
            "a": 0,
            "b": [0.5783, 1, 2.5789],
            "c": {0, 1.87559, 2},
            "d": ["0", "1", "2"],
            "e": {"0", "1", "2"},
            "f": {"a": 0, "b": {"a": 1}},
        }
        self.idk2 = [
            "a",
            "b",
            "c",
            0,
            0.18752,
            2,
            ["a", "b", "c", 0.8579, 1, 2.59757],
            (self.idk, self.idk),
        ]


class MyClassGetState(SlotPickleMixin):
    __slots__ = ["idk", "idk2"]

    def __init__(self):
        self.idk = {
            "a": 0,
            "b": [0.5783, 1, 2.5789],
            "c": {0, 1.87559, 2},
            "d": ["0", "1", "2"],
            "e": {"0", "1", "2"},
            "f": {"a": 0, "b": {"a": 1}},
        }
        self.idk2 = [
            "a",
            "b",
            "c",
            0,
            0.18752,
            2,
            ["a", "b", "c", 0.8579, 1, 2.59757],
            (self.idk, self.idk),
        ]

    def __getstate__(self):
        return SlotPickleMixin.__getstate__(self)

    def __setstate__(self, object_state):
        SlotPickleMixin.__setstate__(self, object_state)


class MyClassSimple(object):
    def __init__(self):
        self.idk = {"a": 0}
        self.idk2 = ["a", 0]


HOMOGENOUS_COMPLEX_DICT = {
    "a": 0,
    "b": 4.2,
    "c": {6.9: "d"},
    -1: {5: 5},
    6: {8.42: -2.5},
}
HOMOGENOUS_COMPLEX_LIST = [
    "a",
    0,
    "b",
    4.2,
    "c",
    [6.9, "d"],
    -1,
    [5, 5],
    6,
    [8.42, -2.5],
]
HOMOGENOUS_COMPLEX_TUPLE = (
    "a",
    0,
    "b",
    4.2,
    "c",
    (6.9, "d"),
    -1,
    (5, 5),
    6,
    (8.42, -2.5),
)
# can't put a set inside a set :(
HOMOGENOUS_COMPLEX_SET = {
    "a",
    0,
    "b",
    4.2,
    "c",
    (6.9, "d"),
    -1,
    (5, 5),
    6,
    (8.42, -2.5),
}

HETEROGENOUS_COMPLEX_DICT = {
    "a": (0, "1"),
    2.53: [3, 1],
    "c": {4.2, 6.9},
    "d": {1: 5.432},
}
HETEROGENOUS_COMPLEX_LIST = [
    "a",
    (0, "1"),
    2.53,
    [3, 1],
    "c",
    {4.2, 6.9},
    "d",
    {1: 5.432},
]
HETEROGENOUS_COMPLEX_TUPLE = (
    "a",
    (0, "1"),
    2.53,
    [3, 1],
    "c",
    {4.2, 6.9},
    "d",
    {1: 5.432},
)
HETEROGENOUS_COMPLEX_SET = {
    "a",
    (0, "1"),
    2.53,
    (3, 1),
    "c",
    (4.2, 6.9),
    "d",
    (1, 5.432),
}

simple_dict_encoded = jsonpickle.encode({"a": 0, "b": 4.2, 3: 6.9})
simple_list_encoded = jsonpickle.encode(["a", 0, "b", 4.2, 3, 6.9])
simple_tuple_encoded = jsonpickle.encode(("a", 0, "b", 4.2, 3, 6.9))
simple_set_encoded = jsonpickle.encode({"a", 0, "b", 4.2, 3, 6.9})

homogenous_dict_encoded = jsonpickle.encode(HOMOGENOUS_COMPLEX_DICT)
homogenous_list_encoded = jsonpickle.encode(HOMOGENOUS_COMPLEX_LIST)
homogenous_tuple_encoded = jsonpickle.encode(HOMOGENOUS_COMPLEX_TUPLE)
homogenous_set_encoded = jsonpickle.encode(HOMOGENOUS_COMPLEX_SET)

heterogenous_dict_encoded = jsonpickle.encode(HETEROGENOUS_COMPLEX_DICT)
heterogenous_list_encoded = jsonpickle.encode(HETEROGENOUS_COMPLEX_LIST)
heterogenous_tuple_encoded = jsonpickle.encode(HETEROGENOUS_COMPLEX_TUPLE)
heterogenous_set_encoded = jsonpickle.encode(HETEROGENOUS_COMPLEX_SET)

simple_class_encoded = jsonpickle.encode(MyClassSimple())
class_encoded = jsonpickle.encode(MyClass())
state_class_encoded = jsonpickle.encode(MyClassGetState())
###################################################################################

# SIMPLE PRIMITIVE ENCODE/DECODE


def simple_dict_encode(benchmark):
    benchmark(jsonpickle.encode, {"a": 0})


def simple_dict_decode(benchmark):
    benchmark(jsonpickle.decode, simple_dict_encoded)


def simple_list_encode(benchmark):
    benchmark(jsonpickle.encode, ["a", 0])


def simple_list_decode(benchmark):
    benchmark(jsonpickle.decode, simple_list_encoded)


def simple_tuple_encode(benchmark):
    benchmark(jsonpickle.encode, ("a", 0))


def simple_tuple_decode(benchmark):
    benchmark(jsonpickle.decode, simple_tuple_encoded)


def simple_set_encode(benchmark):
    benchmark(jsonpickle.encode, {"a", 0})


def simple_set_decode(benchmark):
    benchmark(jsonpickle.decode, simple_set_encoded)


# COMPLEX HOMOGENOUS PRIMITIVE ENCODE/DECODE
def complex_homogenous_dict_encode(benchmark):
    benchmark(jsonpickle.encode, HOMOGENOUS_COMPLEX_DICT)


def complex_homogenous_dict_decode(benchmark):
    benchmark(jsonpickle.decode, homogenous_dict_encoded)


def complex_homogenous_list_encode(benchmark):
    benchmark(jsonpickle.encode, HOMOGENOUS_COMPLEX_LIST)


def complex_homogenous_list_decode(benchmark):
    benchmark(jsonpickle.decode, homogenous_list_encoded)


def complex_homogenous_tuple_encode(benchmark):
    benchmark(jsonpickle.encode, HOMOGENOUS_COMPLEX_TUPLE)


def complex_homogenous_tuple_decode(benchmark):
    benchmark(jsonpickle.decode, homogenous_tuple_encoded)


def complex_homogenous_set_encode(benchmark):
    benchmark(jsonpickle.encode, HOMOGENOUS_COMPLEX_SET)


def complex_homogenous_set_decode(benchmark):
    benchmark(jsonpickle.decode, homogenous_set_encoded)


# COMPLEX HETEROGENOUS PRIMITIVE ENCODE/DECODE
def complex_heterogenous_dict_encode(benchmark):
    benchmark(jsonpickle.encode, HETEROGENOUS_COMPLEX_DICT)


def complex_heterogenous_dict_decode(benchmark):
    benchmark(jsonpickle.decode, heterogenous_dict_encoded)


def complex_heterogenous_list_encode(benchmark):
    benchmark(jsonpickle.encode, HETEROGENOUS_COMPLEX_LIST)


def complex_heterogenous_list_decode(benchmark):
    benchmark(jsonpickle.decode, heterogenous_list_encoded)


def complex_heterogenous_tuple_encode(benchmark):
    benchmark(jsonpickle.encode, HETEROGENOUS_COMPLEX_TUPLE)


def complex_heterogenous_tuple_decode(benchmark):
    benchmark(jsonpickle.decode, heterogenous_tuple_encoded)


def complex_heterogenous_set_encode(benchmark):
    benchmark(jsonpickle.encode, HETEROGENOUS_COMPLEX_SET)


def complex_heterogenous_set_decode(benchmark):
    benchmark(jsonpickle.decode, heterogenous_set_encoded)


# CUSTOM CLASS ENCODE/DECODE
def simple_class_encode(benchmark):
    benchmark(jsonpickle.encode, MyClassSimple())


def simple_class_decode(benchmark):
    benchmark(jsonpickle.decode, simple_class_encoded)


def complex_class_encode(benchmark):
    benchmark(jsonpickle.encode, MyClass())


def complex_class_decode(benchmark):
    benchmark(jsonpickle.decode, class_encoded)


def state_class_encode(benchmark):
    benchmark(jsonpickle.encode, MyClassGetState())


def state_class_decode(benchmark):
    benchmark(jsonpickle.decode, state_class_encoded)



collect_ignore = ['contrib', 'examples']


def pytest_addoption(parser):
    parser.addoption(
        '--repeat', action='store', help='Number of times to repeat each test'
    )


def pytest_generate_tests(metafunc):
    if metafunc.config.option.repeat is not None:
        count = int(metafunc.config.option.repeat)

        # We're going to duplicate these tests by parametrizing them,
        # which requires that each test has a fixture to accept the parameter.
        # We can add a new fixture like so:
        metafunc.fixturenames.append('tmp_ct')

        # Now we parametrize. This is what happens when we do e.g.,
        # @pytest.mark.parametrize('tmp_ct', range(count))
        # def test_foo(): pass
        metafunc.parametrize('tmp_ct', range(count))



"""Test serializing pymongo bson structures"""

import datetime
import json
import pickle

from helper import SkippableTest

import jsonpickle

bson = None


class Object:
    def __init__(self, offset):
        self.offset = datetime.timedelta(offset)

    def __getinitargs__(self):
        return (self.offset,)


class BSONTestCase(SkippableTest):
    def setUp(self):
        global bson
        try:
            bson = __import__('bson.tz_util')
            self.should_skip = False
        except ImportError:
            self.should_skip = True

    def test_FixedOffsetSerializable(self):
        if self.should_skip:
            return self.skip('bson is not installed')
        fo = bson.tz_util.FixedOffset(-60 * 5, 'EST')
        serialized = jsonpickle.dumps(fo)
        restored = jsonpickle.loads(serialized)
        assert vars(restored) == vars(fo)

    def test_timedelta(self):
        if self.should_skip:
            return self.skip('bson is not installed')
        td = datetime.timedelta(-1, 68400)
        serialized = jsonpickle.dumps(td)
        restored = jsonpickle.loads(serialized)
        assert restored == td

    def test_stdlib_pickle(self):
        if self.should_skip:
            return self.skip('bson is not installed')
        fo = bson.tz_util.FixedOffset(-60 * 5, 'EST')
        serialized = pickle.dumps(fo)
        restored = pickle.loads(serialized)
        assert vars(restored) == vars(fo)

    def test_nested_objects(self):
        if self.should_skip:
            return self.skip('bson is not installed')
        o = Object(99)
        serialized = jsonpickle.dumps(o)
        restored = jsonpickle.loads(serialized)
        assert restored.offset == datetime.timedelta(99)

    def test_datetime_with_fixed_offset(self):
        if self.should_skip:
            return self.skip('bson is not installed')
        fo = bson.tz_util.FixedOffset(-60 * 5, 'EST')
        dt = datetime.datetime.now().replace(tzinfo=fo)
        serialized = jsonpickle.dumps(dt)
        restored = jsonpickle.loads(serialized)
        assert restored == dt

    def test_datetime_with_fixed_offset_incremental(self):
        """Test creating an Unpickler and incrementally encoding"""
        if self.should_skip:
            return self.skip('bson is not installed')
        obj = datetime.datetime(2019, 1, 29, 18, 9, 8, 826000, tzinfo=bson.tz_util.utc)
        doc = jsonpickle.dumps(obj)
        # Restore the json using a custom unpickler context.
        unpickler = jsonpickle.unpickler.Unpickler()
        jsonpickle.loads(doc, context=unpickler)
        # Incrementally restore using the same context
        clone = json.loads(doc, object_hook=lambda x: unpickler.restore(x, reset=False))
        assert obj.tzinfo.__reduce__() == clone.tzinfo.__reduce__()



import os
import sys


def ensure_no_files_overwritten(expected_contents):
    # ensure we don't overwrite the user's files, if they run it on their machine
    if os.path.exists("example.json"):
        with open("example.json", "r") as f:
            contents = f.read()
        if contents != expected_contents:
            # exit with a non-zero error code
            sys.exit(
                "I don't want to overwrite the example.json file in my current directory! Please delete example.txt and try again!"
            )



"""
This example demonstrates how to add a custom handler to serialize
Qt's QPointF class jsonpickle using qtpy.

"""

import sys
import unittest

from qtpy import QtCore

import jsonpickle
from jsonpickle import handlers

text_type = eval('unicode') if str is bytes else str


class QReduceHandler(handlers.BaseHandler):
    def flatten(self, obj, data):
        pickler = self.context
        if not pickler.unpicklable:
            return text_type(obj)
        flatten = pickler.flatten
        data['__reduce__'] = [flatten(i, reset=False) for i in obj.__reduce__()[1]]
        return data

    def restore(self, data):
        unpickler = self.context
        restore = unpickler.restore
        reduced = [restore(i, reset=False) for i in data['__reduce__']]
        modulename = reduced[0]
        classname = reduced[1]
        args = reduced[2]
        __import__(modulename)
        module = sys.modules[modulename]
        factory = getattr(module, classname)

        return factory(*args)


handlers.register(QtCore.QPointF, QReduceHandler)


class QtTestCase(unittest.TestCase):
    def test_QPointF_roundtrip(self):
        expect = QtCore.QPointF(1.0, 2.0)
        json = jsonpickle.encode(expect)
        actual = jsonpickle.decode(json)
        self.assertEqual(expect, actual)



import sys
from json import JSONDecodeError

import atheris
from utils import is_expected_error

with atheris.instrument_imports():
    import jsonpickle

# Keys are filenames and values are lists of tuples.
# Each tuple contains a substring of an expected exception message and a line number
# (or -1 to match any line number).
expected_errors = {
    "jsonpickle/unpickler.py": [
        ("object has no attribute 'split'", 283),
        ("not enough values to unpack", 283),
        ("object has no attribute 'split'", 197),
        ("object has no attribute 'split'", 307),
        ("object has no attribute 'encode'", 404),
        ("object has no attribute 'encode'", 407),
        ("object has no attribute 'encode'", 431),
        ("object has no attribute 'encode'", 434),
        ("object is not iterable", 448),
        ("list indices must be integers or slices", 539),
        ("object is not iterable", 448),
        ("object is not iterable", 475),
        ("object is not iterable", 481),
        ("object is not iterable", 508),
        ("object is not iterable", 841),
        ("object is not iterable", 808),
        ("object is not iterable", 859),
        ("object is not iterable", 881),
        ("object is not iterable", 892),
        ("unhashable type", 407),
        ("unhashable type", 841),
        ("object argument after * must be an iterable", 493),
        ("object argument after * must be an iterable", 520),
        ("list index out of range", 160),
        ("'utf-8' codec can't encode character", 407),
        ("too many values to unpack", 283),
        ("too many values to unpack", 511),
        ("object is not callable", 493),
        ("object is not callable", 520),
    ],
    "base64.py": [
        ("bad base85 character at", -1),
        ("base85 overflow in hunk", -1),
        ("Invalid base64-encoded string", -1),
        ("Incorrect padding", -1),
    ],
}


def TestOneInput(data):
    fdp = atheris.FuzzedDataProvider(data)
    fuzz_string = fdp.ConsumeUnicodeNoSurrogates(
        fdp.ConsumeIntInRange(0, fdp.remaining_bytes())
    )

    try:
        jsonpickle.unpickler.decode(fuzz_string)
    except (RecursionError, JSONDecodeError):
        return -1
    except Exception as e:
        if is_expected_error(e, expected_errors):
            return 0
        raise e


def main():
    atheris.Setup(sys.argv, TestOneInput)
    atheris.Fuzz()


if __name__ == "__main__":
    main()



import pytest

try:
    import numpy as np
    import sklearn.datasets
    from sklearn.linear_model import LogisticRegression
    from sklearn.tree import DecisionTreeClassifier
except ImportError:
    pytest.skip('sklearn is not available', allow_module_level=True)

import jsonpickle
import jsonpickle.ext.numpy


@pytest.fixture(scope='module', autouse=True)
def numpy_extension():
    """Initialize the numpy extension for this test module"""
    jsonpickle.ext.numpy.register_handlers()
    yield  # control to the test function.
    jsonpickle.ext.numpy.unregister_handlers()


def test_decision_tree():
    #  Create data.
    np.random.seed(13)
    x_values = np.random.randint(low=0, high=10, size=12)
    x = x_values.reshape(4, 3)
    y_values = np.random.randint(low=0, high=2, size=4)
    y = y_values.reshape(-1, 1)

    # train model
    classifier = DecisionTreeClassifier(max_depth=1)
    classifier.fit(x, y)

    # freeze and thaw
    pickler = jsonpickle.pickler.Pickler()
    unpickler = jsonpickle.unpickler.Unpickler()
    actual = unpickler.restore(pickler.flatten(classifier))

    assert isinstance(actual, classifier.__class__)
    if hasattr(classifier, 'tree_'):
        assert isinstance(actual.tree_, classifier.tree_.__class__)

    # predict from thawed
    array_values = np.array([1, 2, 3])
    array = array_values.reshape(1, -1)
    prediction = actual.predict(array)
    assert prediction[0] == 1

    assert actual.max_depth == classifier.max_depth
    assert actual.score(x, y) == classifier.score(x, y)
    if hasattr(classifier, 'get_depth'):
        assert actual.get_depth() == classifier.get_depth()


def test_nested_array_serialization():
    iris = sklearn.datasets.load_iris()
    X = iris.data[:, :2]
    y = (iris.target != 0) * 1

    pipe = LogisticRegression(solver="lbfgs")
    pipe.fit(X, y)
    pipe.predict(X)

    # Try serializing/deserializing.
    # Pipe is a LogisticRegression object that contains a nested array
    js = jsonpickle.dumps(pipe)
    pipe2 = jsonpickle.loads(js)
    # coef_ only appears on an instance that has already been fit to something
    assert np.all(pipe.coef_ == pipe2.coef_)



import decimal
from hashlib import md5
from warnings import warn

import pytest
from helper import SkippableTest

import jsonpickle


class Thing:
    def __init__(self, name):
        self.name = name
        self.child = None


class A:
    def __init__(self):
        self.id = md5(str(id(self)).encode()).hexdigest()[:5]  # unique enough hash


class BSlots:
    __slots__ = ["a2", "a1", "a3"]

    def __init__(self):
        self.a2 = A()  # set attribs not in alphabetical order
        self.a1 = A()
        self.a3 = self.a1  # create a reference


SAMPLE_DATA = {'things': [Thing('data')]}


class BackendBase(SkippableTest):
    def _is_installed(self, backend):
        if not jsonpickle.util.is_installed(backend):
            return self.skip('%s not available; please install' % backend)

    def set_backend(self, *args):
        backend = args[0]

        self._is_installed(backend)

        jsonpickle.load_backend(*args)
        jsonpickle.set_preferred_backend(backend)

    def set_preferred_backend(self, backend):
        self._is_installed(backend)
        jsonpickle.set_preferred_backend(backend)

    def tearDown(self):
        # always reset to default backend
        jsonpickle.set_preferred_backend('json')

    def assert_roundtrip(self, json_input):
        expect = SAMPLE_DATA
        actual = jsonpickle.decode(json_input)
        assert expect['things'][0].name == actual['things'][0].name
        assert expect['things'][0].child == actual['things'][0].child

        pickled = jsonpickle.encode(SAMPLE_DATA)
        actual = jsonpickle.decode(pickled)
        assert expect['things'][0].name == actual['things'][0].name
        assert expect['things'][0].child == actual['things'][0].child

    def test_None_dict_key(self):
        """Ensure that backends produce the same result for None dict keys"""
        data = {None: None}
        expect = {'null': None}
        pickle = jsonpickle.encode(data)
        actual = jsonpickle.decode(pickle)
        assert expect == actual

    def test_encode_with_indent_and_separators(self):
        obj = {
            'a': 1,
            'b': 2,
        }
        expect = '{\n' '    "a": 1,\n' '    "b": 2\n' '}'
        actual = jsonpickle.encode(obj, indent=4, separators=(',', ': '))
        assert expect == actual


class JsonTestCase(BackendBase):
    def setUp(self):
        self.set_preferred_backend('json')

    def test_backend(self):
        expected_pickled = (
            '{"things": [{'
            '"py/object": "backend_test.Thing", '
            '"name": "data", '
            '"child": null} '
            ']}'
        )
        self.assert_roundtrip(expected_pickled)


class SimpleJsonTestCase(BackendBase):
    def setUp(self):
        self.set_preferred_backend('simplejson')

    def test_backend(self):
        expected_pickled = (
            '{"things": [{'
            '"py/object": "backend_test.Thing", '
            '"name": "data", '
            '"child": null}'
            ']}'
        )
        self.assert_roundtrip(expected_pickled)

    def test_decimal(self):
        # Default behavior: Decimal is preserved
        obj = decimal.Decimal(0.5)
        as_json = jsonpickle.dumps(obj)
        clone = jsonpickle.loads(as_json)
        assert isinstance(clone, decimal.Decimal)
        assert obj == clone

        # Custom behavior: we want to use simplejson's Decimal support.
        jsonpickle.set_encoder_options('simplejson', use_decimal=True, sort_keys=True)

        jsonpickle.set_decoder_options('simplejson', use_decimal=True)

        # use_decimal mode allows Decimal objects to pass-through to simplejson.
        # The end result is we get a simple '0.5' value as our json string.
        as_json = jsonpickle.dumps(obj, unpicklable=True, use_decimal=True)
        assert as_json == '0.5'
        # But when loading we get back a Decimal.
        clone = jsonpickle.loads(as_json)
        assert isinstance(clone, decimal.Decimal)

        # side-effect: floats become decimals too!
        obj = 0.5
        as_json = jsonpickle.dumps(obj)
        clone = jsonpickle.loads(as_json)
        assert isinstance(clone, decimal.Decimal)
        # options are persisted unless we disable them
        jsonpickle.set_encoder_options('simplejson', use_decimal=False)
        jsonpickle.set_decoder_options('simplejson', use_decimal=False)

    def test_sort_keys(self):
        jsonpickle.set_encoder_options('simplejson', sort_keys=True)
        b = BSlots()
        with pytest.raises(TypeError):
            jsonpickle.encode(b, keys=True, warn=True)
        # return encoder options to default
        jsonpickle.set_encoder_options('simplejson', sort_keys=False)


def has_module(module):
    try:
        __import__(module)
    except ImportError:
        warn(module + ' module not available for testing, ' 'consider installing')
        return False
    return True


class UJsonTestCase(BackendBase):
    def setUp(self):
        self.set_preferred_backend('ujson')

    def test_backend(self):
        expected_pickled = (
            '{"things":[{'
            r'"py\/object":"backend_test.Thing",'
            '"name":"data","child":null}'
            ']}'
        )
        self.assert_roundtrip(expected_pickled)


class YamlTestCase(BackendBase):
    def setUp(self):
        self.set_preferred_backend('yaml')

    def test_backend(self):
        expected_pickled = (
            'things:\n'
            '  - py/object: backend_test.Thing\n'
            '    name: data\n'
            '    child: null\n'
        )
        self.assert_roundtrip(expected_pickled)



"""Wizard tests from petrounias.org

http://www.petrounias.org/articles/2014/09/16/pickling-python-collections-with-non-built-in-type-keys-and-cycles/

Includes functionality to assist with adding compatibility to jsonpickle.

"""

import collections

from jsonpickle import decode, encode


class World:
    def __init__(self):
        self.wizards = []


class Wizard:
    def __init__(self, world, name):
        self.name = name
        self.spells = collections.OrderedDict()
        world.wizards.append(self)

    def __cmp__(self, other):
        for (ka, va), (kb, vb) in zip(self.spells.items(), other.spells.items()):
            cmp_name = cmp(ka.name, kb.name)  # noqa: F821
            if cmp_name != 0:
                print(f'Wizards cmp: {ka.name} != {kb.name}')
                return cmp_name
            for sa, sb in zip(va, vb):
                cmp_spell = cmp(sa, sb)  # noqa: F821
                if cmp_spell != 0:
                    print(f'Spells cmp: {sa.name} != {sb.name}')
                    return cmp_spell
        return cmp(self.name, other.name)  # noqa: F821

    def __eq__(self, other):
        for (ka, va), (kb, vb) in zip(self.spells.items(), other.spells.items()):
            if ka.name != kb.name:
                print(f'Wizards differ: {ka.name} != {kb.name}')
                return False
            for sa, sb in zip(va, vb):
                if sa != sb:
                    print(f'Spells differ: {sa.name} != {sb.name}')
                    return False
        return self.name == other.name

    def __hash__(self):
        return hash('Wizard %s' % self.name)


class Spell:
    def __init__(self, caster, target, name):
        self.caster = caster
        self.target = target
        self.name = name
        try:
            spells = caster.spells[target]
        except KeyError:
            spells = caster.spells[target] = []
        spells.append(self)

    def __cmp__(self, other):
        return (
            cmp(self.name, other.name)  # noqa: F821
            or cmp(self.caster.name, other.caster.name)  # noqa: F821
            or cmp(self.target.name, other.target.name)  # noqa: F821
        )  # noqa: F821

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.caster.name == other.caster.name
            and self.target.name == other.target.name
        )

    def __hash__(self):
        return hash(f'Spell {self.name} by {self.caster.name} on {self.target.name}')


def hashsum(items):
    return sum([hash(x) for x in items])


def compare_spells(a, b):
    for (ka, va), (kb, vb) in zip(a.items(), b.items()):
        if ka != kb:
            print(f'Keys differ: {ka} != {kb}')
            return False
    return True


def test_without_pickling():
    world = World()
    wizard_merlin = Wizard(world, 'Merlin')
    wizard_morgana = Wizard(world, 'Morgana')
    spell_a = Spell(wizard_merlin, wizard_morgana, 'magic-missile')
    spell_b = Spell(wizard_merlin, wizard_merlin, 'stone-skin')
    spell_c = Spell(wizard_morgana, wizard_merlin, 'geas')
    assert wizard_merlin.spells[wizard_morgana][0] == spell_a
    assert wizard_merlin.spells[wizard_merlin][0] == spell_b
    assert wizard_morgana.spells[wizard_merlin][0] == spell_c
    # Merlin has cast Magic Missile on Morgana, and Stone Skin on himself
    assert wizard_merlin.spells[wizard_morgana][0].name == 'magic-missile'
    assert wizard_merlin.spells[wizard_merlin][0].name == 'stone-skin'
    # Morgana has cast Geas on Merlin
    assert wizard_morgana.spells[wizard_merlin][0].name == 'geas'
    # Merlin's first target was Morgana
    merlin_spells = wizard_merlin.spells
    merlin_spells_keys = list(merlin_spells.keys())
    assert merlin_spells_keys[0] in wizard_merlin.spells
    assert merlin_spells_keys[0] == wizard_morgana
    # Merlin's second target was himself
    assert merlin_spells_keys[1] in wizard_merlin.spells
    assert merlin_spells_keys[1] == wizard_merlin
    # Morgana's first target was Merlin
    morgana_spells_keys = list(wizard_morgana.spells.keys())
    assert morgana_spells_keys[0] in wizard_morgana.spells
    assert morgana_spells_keys[0] == wizard_merlin
    # Merlin's first spell cast with himself as target is in the dictionary,
    # first by looking up directly with Merlin's instance object...
    assert wizard_merlin == wizard_merlin.spells[wizard_merlin][0].target
    # ...and then with the instance object directly from the dictionary keys
    assert wizard_merlin == merlin_spells[merlin_spells_keys[1]][0].target
    # Ensure Merlin's object is unique...
    assert id(wizard_merlin) == id(merlin_spells_keys[1])
    # ...and consistently hashed
    assert hash(wizard_merlin) == hash(merlin_spells_keys[1])


def test_with_pickling():
    world = World()
    wizard_merlin = Wizard(world, 'Merlin')
    wizard_morgana = Wizard(world, 'Morgana')
    wizard_morgana_prime = Wizard(world, 'Morgana')
    assert wizard_morgana.__dict__ == wizard_morgana_prime.__dict__

    spell_a = Spell(wizard_merlin, wizard_morgana, 'magic-missile')
    spell_b = Spell(wizard_merlin, wizard_merlin, 'stone-skin')
    spell_c = Spell(wizard_morgana, wizard_merlin, 'geas')
    assert wizard_merlin.spells[wizard_morgana][0] == spell_a
    assert wizard_merlin.spells[wizard_merlin][0] == spell_b
    assert wizard_morgana.spells[wizard_merlin][0] == spell_c

    flat_world = encode(world, keys=True)
    u_world = decode(flat_world, keys=True)
    u_wizard_merlin = u_world.wizards[0]
    u_wizard_morgana = u_world.wizards[1]
    morgana_spells_encoded = encode(wizard_morgana.spells, keys=True)
    morgana_spells_decoded = decode(morgana_spells_encoded, keys=True)
    assert wizard_morgana.spells == morgana_spells_decoded

    morgana_encoded = encode(wizard_morgana, keys=True)
    morgana_decoded = decode(morgana_encoded, keys=True)
    assert wizard_morgana == morgana_decoded
    assert hash(wizard_morgana) == hash(morgana_decoded)
    assert wizard_morgana.spells == morgana_decoded.spells
    # Merlin has cast Magic Missile on Morgana, and Stone Skin on himself
    merlin_spells = u_wizard_merlin.spells
    assert merlin_spells[u_wizard_morgana][0].name == 'magic-missile'
    assert merlin_spells[u_wizard_merlin][0].name == 'stone-skin'
    # Morgana has cast Geas on Merlin
    assert u_wizard_morgana.spells[u_wizard_merlin][0].name == 'geas'
    # Merlin's first target was Morgana
    merlin_spells_keys = list(u_wizard_merlin.spells.keys())
    assert merlin_spells_keys[0] in u_wizard_merlin.spells
    assert merlin_spells_keys[0] == u_wizard_morgana
    # Merlin's second target was himself
    assert merlin_spells_keys[1] in u_wizard_merlin.spells
    assert merlin_spells_keys[1] == u_wizard_merlin
    # Morgana's first target was Merlin
    morgana_spells_keys = list(u_wizard_morgana.spells.keys())
    assert morgana_spells_keys[0] in u_wizard_morgana.spells
    assert morgana_spells_keys[0] == u_wizard_merlin
    # Merlin's first spell cast with himself as target is in the dict.
    # First try the lookup with Merlin's instance object
    assert u_wizard_merlin == merlin_spells[u_wizard_merlin][0].target
    # Next try the lookup with the object from the dictionary keys.
    assert u_wizard_merlin == merlin_spells[merlin_spells_keys[1]][0].target
    # Ensure Merlin's object is unique and consistently hashed.
    assert id(u_wizard_merlin) == id(merlin_spells_keys[1])
    assert hash(u_wizard_merlin) == hash(merlin_spells_keys[1])



# Copyright (C) 2008 John Paulett (john -at- paulett.org)
# Copyright (C) 2009-2018 David Aguilar (davvid -at- gmail.com)
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

import io
import time

from jsonpickle import util


class Thing:
    def __init__(self, name):
        self.name = name
        self.child = None


class DictSubclass(dict):
    pass


class ListSubclass(list):
    pass


class MethodTestClass:
    variable = None

    @staticmethod
    def static_method():
        pass

    @classmethod
    def class_method(cls):
        pass

    def bound_method(self):
        pass


class MethodTestSubclass(MethodTestClass):
    pass


class MethodTestOldStyle:
    def bound_method(self):
        pass


def test_is_primitive_int():
    assert util.is_primitive(0)
    assert util.is_primitive(3)
    assert util.is_primitive(-3)


def test_is_primitive_float():
    assert util.is_primitive(0)
    assert util.is_primitive(3.5)
    assert util.is_primitive(-3.5)
    assert util.is_primitive(float(3))


def test_is_primitive_long():
    assert util.is_primitive(2**64)


def test_is_primitive_bool():
    assert util.is_primitive(True)
    assert util.is_primitive(False)


def test_is_primitive_None():
    assert util.is_primitive(None)


def test_is_primitive_bytes():
    assert not util.is_primitive(b'hello')
    assert not util.is_primitive(b'foo')
    assert util.is_primitive('foo')


def test_is_primitive_unicode():
    assert util.is_primitive('')
    assert util.is_primitive('hello')


def test_is_primitive_list():
    assert not util.is_primitive([])
    assert not util.is_primitive([4, 4])


def test_is_primitive_dict():
    assert not util.is_primitive({'key': 'value'})
    assert not util.is_primitive({})


def test_is_primitive_tuple():
    assert not util.is_primitive((1, 3))
    assert not util.is_primitive((1,))


def test_is_primitive_set():
    assert not util.is_primitive({1, 3})


def test_is_primitive_object():
    assert not util.is_primitive(Thing('test'))


def test_is_list_list():
    assert util.is_list([1, 2])


def test_is_list_set():
    assert util.is_set({1, 2})


def test_is_list_tuple():
    assert util.is_tuple((1, 2))


def test_is_list_dict():
    assert not util.is_list({'key': 'value'})
    assert not util.is_set({'key': 'value'})
    assert not util.is_tuple({'key': 'value'})


def test_is_list_other():
    assert not util.is_list(1)
    assert not util.is_set(1)
    assert not util.is_tuple(1)


def test_is_sequence_various():
    assert util.is_sequence([])
    assert util.is_sequence(tuple())
    assert util.is_sequence(set())


def test_is_dictionary_dict():
    assert util.is_dictionary({})


def test_is_dicitonary_sequences():
    assert not util.is_dictionary([])
    assert not util.is_dictionary(set())


def test_is_dictionary_tuple():
    assert not util.is_dictionary(tuple())


def test_is_dictionary_primitive():
    assert not util.is_dictionary(int())
    assert not util.is_dictionary(None)
    assert not util.is_dictionary('')


def test_is_dictionary_subclass_dict():
    assert not util.is_dictionary_subclass({})


def test_is_dictionary_subclass_subclass():
    assert util.is_dictionary_subclass(DictSubclass())


def test_is_sequence_subclass_subclass():
    assert util.is_sequence_subclass(ListSubclass())


def test_is_sequence_subclass_list():
    assert not util.is_sequence_subclass([])


def test_is_noncomplex_time_struct():
    t = time.struct_time('123456789')
    assert util.is_noncomplex(t)


def test_is_noncomplex_other():
    assert not util.is_noncomplex('a')


def test_is_function_builtins():
    assert util.is_function(globals)


def test_is_function_lambda():
    assert util.is_function(lambda: False)


def test_is_function_instance_method():
    class Foo:
        def method(self):
            pass

        @staticmethod
        def staticmethod():
            pass

        @classmethod
        def classmethod(cls):
            pass

    f = Foo()
    assert util.is_function(f.method)
    assert util.is_function(f.staticmethod)
    assert util.is_function(f.classmethod)


def test_itemgetter():
    expect = '0'
    actual = util.itemgetter((0, 'zero'))
    assert expect == actual


def test_has_method():
    instance = MethodTestClass()
    x = 1
    has_method = util.has_method
    # no attribute
    assert not has_method(instance, 'foo')
    # builtin method type
    assert not has_method(int, '__getnewargs__')
    assert has_method(x, '__getnewargs__')
    # staticmethod
    assert has_method(instance, 'static_method')
    assert has_method(MethodTestClass, 'static_method')
    # not a method
    assert not has_method(instance, 'variable')
    assert not has_method(MethodTestClass, 'variable')
    # classmethod
    assert has_method(instance, 'class_method')
    assert has_method(MethodTestClass, 'class_method')
    # bound method
    assert has_method(instance, 'bound_method')
    assert not has_method(MethodTestClass, 'bound_method')
    # subclass bound method
    sub_instance = MethodTestSubclass()
    assert has_method(sub_instance, 'bound_method')
    assert not has_method(MethodTestSubclass, 'bound_method')
    # old style object
    old_instance = MethodTestOldStyle()
    assert has_method(old_instance, 'bound_method')
    assert not has_method(MethodTestOldStyle, 'bound_method')


def test_importable_name():
    func_being_tested_obj = util.importable_name
    io_method_obj = io.BytesIO(b'bytes').readline
    assert (
        util.importable_name(func_being_tested_obj) == 'jsonpickle.util.importable_name'
    )
    assert util.importable_name(io_method_obj) == '_io.BytesIO.readline'
    assert util.importable_name(dict) == 'builtins.dict'



import sys

import jsonpickle


def _roundtrip(obj):
    """Verify object equality after encoding and decoding to/from jsonpickle"""
    pickled = jsonpickle.encode(obj)
    unpickled = jsonpickle.decode(pickled)
    assert obj == unpickled


def test_zoneinfo():
    """zoneinfo objects can roundtrip"""
    if sys.version_info < (3, 9):
        return
    from zoneinfo import ZoneInfo

    _roundtrip(ZoneInfo('Australia/Brisbane'))



import csv

# makes math far easier
import numpy as np


def filter_results_by_name(results, substr):
    perfs = []
    for name, perf in results.items():
        if substr in name:
            perfs.append(perf)
    return np.array(perfs)


def main():
    # using dicts seems easier at first but we'd have to eventually convert
    # to a list for easy elementwise division or use an ugly loop
    funcs = []
    main = []
    new = []

    with open("perf.csv", "r") as f:
        csv_reader = csv.reader(f)
        # skip header row
        next(csv_reader)

        # csv is sorted by name so we can assume even rows are main and odd are new
        for indx, (func, median) in enumerate(csv_reader):
            func = func.split("::")[-1]
            # indx is even
            if not (indx % 2):
                funcs.append(func)
                main.append(float(median))
            else:
                new.append(float(median))

    main = np.array(main)
    new = np.array(new)

    relative_perfs = new / main
    # more stats will be included when needed
    print(
        f"The new branch is {round(relative_perfs.mean(), 4)}x faster than the main branch aross all tests!"
    )

    results = {}

    for func_name, relative_perf in zip(funcs, relative_perfs):
        # this allows us to categorize classes of results, such as encode vs decode
        results[func_name] = relative_perf

    encode_perfs = filter_results_by_name(results, "encode")
    decode_perfs = filter_results_by_name(results, "decode")
    print(
        f"The new branch is {round(encode_perfs.mean(), 4)}x faster than the main branch aross encode tests!"
    )
    print(
        f"The new branch is {round(decode_perfs.mean(), 4)}x faster than the main branch aross decode tests!"
    )


if __name__ == "__main__":
    main()



# Copyright (C) 2008 John Paulett (john -at- paulett.org)
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
from helper import SkippableTest

import jsonpickle

RSS_DOC = """<?xml version="1.0" encoding="utf-8"?>
<feed xmlns="http://www.w3.org/2005/Atom"
      xml:base="http://example.org/" xml:lang="en">
  <title type="text">Sample Feed</title>
  <subtitle type="html">For documentation &lt;em&gt;only&lt;/em&gt;</subtitle>
  <link rel="alternate" type="html" href="/"/>
  <link rel="self" type="application/atom+xml"
        href="http://www.example.org/atom10.xml"/>
  <rights type="html">&lt;p>Copyright 2005, Mark Pilgrim&lt;/p>&lt;</rights>

  <generator uri="http://example.org/generator/" version="4.0">
    Sample Toolkit
  </generator>
  <id>tag:feedparser.org,2005-11-09:/docs/examples/atom10.xml</id>
  <updated>2005-11-09T11:56:34Z</updated>
  <entry>
    <title>First entry title</title>
    <link rel="alternate" href="/entry/3"/>
    <link rel="related" type="text/html" href="http://search.example.com/"/>

    <link rel="via" type="text/html"
          href="http://toby.example.com/examples/atom10"/>
    <link rel="enclosure" type="video/mpeg4"
          href="http://www.example.com/movie.mp4" length="42301"/>
    <id>tag:feedparser.org,2005-11-09:/docs/examples/atom10.xml:3</id>
    <published>2005-11-09T00:23:47Z</published>
    <updated>2005-11-09T11:56:34Z</updated>
    <author>
      <name>Mark Pilgrim</name>

      <uri>http://diveintomark.org/</uri>
      <email>mark@example.org</email>
    </author>
    <contributor>
      <name>Joe</name>
      <uri>http://example.org/joe/</uri>
      <email>joe@example.org</email>

    </contributor>
    <contributor>
      <name>Sam</name>
      <uri>http://example.org/sam/</uri>
      <email>sam@example.org</email>
    </contributor>
    <summary type="text">Watch out for nasty tricks</summary>

    <content type="xhtml" xml:base="http://example.org/entry/3"
             xml:lang="en-US">
      <div xmlns="http://www.w3.org/1999/xhtml">
        Watch out for <span style="
            background: url(javascript:window.location='http://example.org/')
        ">
            nasty tricks</span>
        </div>
    </content>
  </entry>
</feed>"""


class FeedParserTestCase(SkippableTest):
    def setUp(self):
        try:
            import feedparser

            self.should_skip = False
            self.doc = feedparser.parse(RSS_DOC)
        except ImportError:
            self.should_skip = True
            return

    def test(self):
        if self.should_skip:
            return self.skip('feedparser module not available, please install')
        pickled = jsonpickle.encode(self.doc)
        unpickled = jsonpickle.decode(pickled)
        assert self.doc['feed']['title'] == unpickled['feed']['title']



"""Test serializing ecdsa keys"""

import pytest
from helper import SkippableTest

import jsonpickle


@pytest.fixture(scope='module', autouse=True)
def gmpy_extension():
    """Initialize the gmpy extension for this test module"""
    try:
        jsonpickle.ext.gmpy.register_handlers()
        yield  # control to the test function.
        jsonpickle.ext.gmpy.unregister_handlers()
    except AttributeError:
        pytest.skip(
            "gmpy was not detected, please try installing it for more complete tests!"
        )


class EcdsaTestCase(SkippableTest):
    def setUp(self):
        try:
            from ecdsa import NIST384p
            from ecdsa.keys import SigningKey

            self.NIST384p = NIST384p
            self.SigningKey = SigningKey
            self.should_skip = False
        except ImportError:
            self.should_skip = True

    def test_roundtrip(self):
        if self.should_skip:
            return self.skip('ecdsa module is not installed')

        message = b'test'
        key_pair = self.SigningKey.generate(curve=self.NIST384p)
        sig = key_pair.sign(message)
        serialized = jsonpickle.dumps(key_pair.get_verifying_key())
        restored = jsonpickle.loads(serialized)
        assert restored.verify(sig, message)
